import numpy as np
from trdstest.ops import TR_initcoreten, Z_neq, tenmat_sb
from trdstest.functions import TR_initcoreten, Msum_fun, Gfold, Gunfold, Pro2TraceNorm, coreten2tr, mytenmat

def TRLRF(data, W, r, maxiter=300, K=1e0, ro=1.1, Lambda=5, tol=1e-6):
    T = data * W
    N = T.ndim
    S = T.shape
    S_1, S_2, S_3 = S
    X = np.random.rand(S_1, S_2, S_3)

    G = TR_initcoreten(S, r)

    M = np.zeros((N, 3), dtype=np.object)
    Y = np.zeros((N, 3), dtype=np.object)

    for i in range(N):
        G[i] = 1 * G[i]
        for j in range(3):
            M[i][j] = np.zeros(G[i].shape)
            Y[i][j] = np.sign(G[i])

    K_max = 10 ** 2
    Convergence_rec = np.zeros(maxiter)

    iter = 0
    while iter < maxiter:

        # update G
        for n in range(N):
            Msum = Msum_fun(M)
            Ysum = Msum_fun(Y)

            Q = tenmat_sb(Z_neq(G, n), 2)
            Q = Q.T

            G[n] = Gfold(np.dot((np.dot((Lambda * tenmat_sb(X, n + 1)), Q.T) + K * Gunfold(
                Msum[n], 1) + Gunfold(Ysum[n], 1)), np.linalg.pinv(
                (Lambda * (np.dot(Q, Q.T)) + 3 * K * np.eye(Q.shape[0], Q.shape[0])))), G[n].shape, 1)

            # update  M

            for j in range(3):
                Df = Gunfold(G[n] - Y[n][j] / K, j)
                M[n][j] = Gfold(Pro2TraceNorm(Df, 1 / K)[0], G[n].shape, j)

        # update X
        lastX = X
        X_hat = coreten2tr(G)
        X = X_hat
        X[W == 1] = T[W == 1]

        # update Y
        for n in range(N):
            for j in range(0, 3):
                Y[n, j] = Y[n, j] + K * (M[n, j] - G[n])
        K = min(K * ro, K_max)

        # evaluation
        G_out = G

        err_x = np.abs(
            (np.linalg.norm(lastX.T.flatten()) - np.linalg.norm(X.T.flatten())) / np.linalg.norm(X.T.flatten()))

        if err_x < tol:
            print('iteration stop at %f\n' % iter)
            break
        if iter % 100 == 0 or iter == 0:
            Ssum_G = 0  # singular value
            for j in range(N):
                _, vz1, __ = np.linalg.svd(mytenmat(G[0], 1, 1))
                _, vz2, __ = np.linalg.svd(mytenmat(G[0], 2, 1))
                _, vz3, __ = np.linalg.svd(mytenmat(G[0], 3, 1))

                Ssum_G = Ssum_G + np.sum(vz1.T.flatten()) + np.sum(vz2.T.flatten()) + np.sum(vz3.T.flatten())
            f_left = Ssum_G
            f_right = Lambda * (np.linalg.norm(X.T.flatten() - X_hat.T.flatten())) ** 2
            Convergence_rec[iter] = f_left + f_right
            print('TRLRF: Iter %f, Diff %d, Reg %d, Fitting %d' % (iter, err_x, f_left, f_right))
        iter = iter + 1

    return X, G_out, Convergence_rec


import numpy as np

def unfolding(T, n=0):
    """Mode-n unfolding of a tensor
    Parameters
    ----------
    T : ndarray
        Tensor to be unfolded
    n : int
        mode n to split
    Returns
    -------
    unfolded_T : ndarray
        unfolded tensor
    """
    T_= np.moveaxis(T, n, 0)
    unfolded_T = np.reshape(T_, (T.shape[n], -1))
    return unfolded_T

def folding(unfolded_T, n, shape):
    """Refolds the Mode-n unfolded tensor
    Parameters
    ----------
    unfolded_T : ndarray
    n : int
    shape : tuple or list
    Returns
    -------
    folded_T : ndarray
        Folded tensor
    """
    if type(shape) == tuple:
        T_shape = list(shape)
    else:
        T_shape = shape
    mode_dim = T_shape.pop(n)
    T_shape.insert(0, mode_dim)
    T_ = np.reshape(unfolded_T, T_shape)
    folded_T = np.moveaxis(T_, 0, n)
    return folded_T

def tensor2vec(T):
    """Vectorization of a tensor
    Parameters
    ----------
    T : ndarray
    Returns
    -------
    V : 1D-array
        vectorised tensor
    """
    V = np.ravel(T)
    return V

def vec2tensor(V, shape):
    """reshape the vectorised tensor to tensor
    Parameters
    ----------
    V : 1D-array
        vectorised tensor
    shape : tuple
        shape of tensor
    Returns
    -------
    T : ndarray
        tesnorized tensor
    """
    T = np.reshape(V, shape)
    return T

def tensor2mat(T, n):
    """Matricize a tensor
    Parameters
    ----------
    T : ndarray
    n : int
    Returns
    -------
    M : 2D-array
        Matricization of a tensor
    """
    M_ = T.transpose(np.append(np.arange(n - 1, T.ndim), np.arange(0, n - 1)))
    M = M_.reshape(T.shape[n-1], int(T.size/T.shape[n-1]), order='F').copy()
    # print("Type: %d" %(mat_type), ", Reshape at mode-%d" %(n), ", Transpose index:", arr, ", Matrix size: %u x %u" %(mat.shape[0], mat.shape[1]))
    return M

def mat2tensor(M, n, shape):
    """Reshape the "matricized tensor" to tensor
    Parameters
    ----------
    M : 2D-array
    n : int
    shape : tuple
    Returns
    -------
    T : ndarray
        tesnorized tensor
    """
    arr = np.append(int(np.prod(shape[n-1:len(shape)])), int(np.prod(shape[0:n - 1])))
    T_ = M.reshape(arr, order = 'F').transpose(1, 0).copy()
    T = T_.reshape(shape, order = 'F').copy()
    # print("The size of tensor is", output_tensor.shape)
    return T

# generate tr core tensors randomly
def init_tr_cores(tensor_size, tr_rank):
    tr_cores = []
    ndims = len(tensor_size)
    tr_rank.append(tr_rank[0])
    for n in range(0, ndims):
        tr_cores.append(0.1 * np.random.rand(tr_rank[n], tensor_size[n], tr_rank[n + 1]))
    return tr_cores



import numpy as np
# from decomposition import TRSVD
# from decomposition import TRALS
from decomposition import TRALSAR



T = np.random.rand(3,4,5,6)
ep = 1e-5
tau = 1e-4
epoch = 20
#
# node1, r1 = TRSVD(T, ep=1e-6)
# node2, r2 = TRALS(T, tol=1e-6)
node3, r3 = TRALSAR(T, ep=1e-6)


import numpy as np
from trdstest.base import unfold, fold
from trdstest.ops import Z_neq, tenmat_sb


def n_mode_product(T, M, n):
    """n-mode product of a tensor with a matrix.
    Parameters
    ----------
    T : ndarray
    M : 2D-array
    n : int
        n mode
    Returns
    -------
    product : ndarray
        n-mode product
    """
    product_ = M @ unfold(T, n)
    res_shape = list(T.shape)
    if M.ndim == 1:
        if len(res_shape) > 1:
            res_shape[n] = 1
        else:
            res_shape[1] = 1
        product = fold(product_, n, res_shape).squeeze()
    else:
        product = fold(product_, n, res_shape)
    return product

def inner_product(T1, T2):
    """inner product of two tensor with same shape
    Parameters
    ----------
    T1 : ndarray
    T2 : ndarray
    Returns
    -------
    product : ndarray
        n-mode product
    """
    product = np.sum(T1 * T2)
    return product

def n_mode_inner_product(T1, T2, n):
    """n mode inner product of two tensor
    Parameters
    ----------
    T1 : ndarray
    T2 : ndarray
    n : int
        n mode
    Returns
    -------
    product : ndarray
        n-mode product
    """
    common_size = int(np.prod(list(T1.shape)[-n:]))
    shape = list(T1.shape)[:-n] + list(T2.shape)[n:]
    product_ = np.reshape(T1, (-1, common_size)) @ np.reshape(T2, (common_size, -1))
    product = product_.reshape(shape)
    return product

def outer_products(Ts):
    """outer product of a list of tensors
    Parameters
    ----------
    Ts : list or tuple
        a list of tensors
    Returns
    -------
    product : ndarray
        outer_products
    """
    for i in range(len(Ts)):
        if i == 0:
            product = Ts[0]
            continue
        shape_before = product.shape + (1,) * Ts[i].ndim
        shape_after = (1,) * product.ndim + Ts[i].shape
        product = product.reshape(shape_before) * Ts[i].reshape(shape_after)
    return product


def kronecker(Ms):
    """Kronecker product of a list of matrices
    Parameters
    ----------
    Ms : list or tuple
        a list of matrices
    Returns
    -------
    product : ndarray
        kronecker products
    """
    for i, M in enumerate(Ms[::1]):
        if not i:
            product = M
        else:
            product = np.kron(product, M)
    return product


def khatri_rao(Ms):
    """Khatri-Rao product of a list of matrices
    ----------
    matrices : 2D-array list
        list of matrices with the same number of columns, i.e.::
            for i in len(matrices):
                matrices[i].shape = (n_i, m)
    reverse : bool, optional
        if True, the order of the matrices is reversed
    Returns
    -------
    khatri_rao_product: matrix of shape ``(prod(n_i), m)``
        where ``prod(n_i) = prod([m.shape[0] for m in matrices])``
        i.e. the product of the number of rows of all the matrices in the product.
    """

    n_columns = Ms[0].shape[1]
    n_factors = len(Ms)

    start = ord('a')
    common_dim = 'z'
    target = ''.join(chr(start + i) for i in range(n_factors))
    source = ','.join(i + common_dim for i in target)
    operation = source + '->' + target + common_dim
    return np.einsum(operation, *Ms).reshape((-1, n_columns))

# TRLRF
def coreten2tr(Z):
    N = Z.size
    S = np.zeros(3, dtype=object)
    for i in range(N):
        S[i] = Z[i].shape[1]

    P = Z[0]
    for i in range(1, N):
        L = np.reshape(P, (int(P.size / Z[i - 1].shape[2]), Z[i - 1].shape[2]))
        R = np.reshape(Z[i], (Z[i].shape[0], int(S[i] * Z[i].shape[2])))

        P = np.dot(L, R)

    P = np.reshape(P, (Z[0].shape[0], np.prod(S), Z[N - 1].shape[2]))
    P = np.transpose(P, [1, 2, 0])
    P = np.reshape(P, (np.prod(S), int(Z[0].shape[0] * Z[0].shape[0])))

    temp = np.eye(Z[0].shape[0], Z[0].shape[0])

    P = np.dot(P, (temp.T.flatten()))
    X = np.reshape(P, S)

    return X

# generate tensor W by size and missing rate
def gen_W(S, mr):
    # np.prod(S)

    Omega_1 = np.random.permutation(np.prod(S))
    Omega = Omega_1[0:int(round((1 - mr) * np.prod(S)))].tolist()
    W_1 = np.zeros(np.prod(S))

    W_1[Omega] = 1
    W = W_1.reshape(S)

    return W

# fold core tensor to original size
def Gfold(Gm, SGt, n):
    if n == 0:
        Gt_out = np.reshape(Gm, (SGt[0], SGt[1], SGt[2]))
    elif n == 1:
        Gt_out = np.transpose(np.reshape(Gm, (SGt[1], SGt[0], SGt[2])), (1, 0, 2))
    elif n == 2:
        Gt_out = np.transpose(np.reshape(Gm, (SGt[2], SGt[0], SGt[1])), (1, 2, 0))
    else:
        print('wrong!')

    return Gt_out


# only for unfold core tensors to 3 different mode
def Gunfold(Gt, n):
    if n == 0:
        Gm = np.reshape(Gt, (Gt.shape[0], Gt.shape[1] * Gt.shape[2]))
    elif n == 1:
        Gm = np.reshape(np.transpose(Gt, (1, 0, 2)), (Gt.shape[1], Gt.shape[0] * Gt.shape[2]))
    elif n == 2:
        Gm = np.reshape(np.transpose(Gt, (2, 0, 1)), (Gt.shape[2], Gt.shape[0] * Gt.shape[1]))
    else:
        print('wrong!')
    return Gm


def Msum_fun(M):
    N = np.size(M, 0)
    Msum_out = np.zeros((N,), dtype=np.object)
    for i in range(N):
        Msum_out[i] = M[i][0] + M[i][1] + M[i][2]

    return Msum_out


# from tensor to matrix
# 1 kolda mat, 2 tt mat, 3 tr mat
def mytenmat(X, n, type):
    S = X.shape
    N = len(S)
    if type == 1:  # the In x I1...In-1In+1...IN type
        if n == 1:
            Xn = np.reshape(X, (S[0], np.prod(S[1:])))
        elif n == N:
            Xn = np.transpose(np.reshape(X, (np.prod(S[0:-1]), S[N - 1])), (1, 0))
        else:
            arr_1 = np.array(n - 1)
            arr_2 = np.arange(0, n - 1)
            arr_3 = np.arange(n, N)
            arr = np.append(np.append(arr_1, arr_2), arr_3)
            Xn = np.reshape(np.transpose(X, arr), (S[n - 1], int(np.prod(S) / S[n - 1])))

    elif type == 2:  # the I1...In x In+1...IN type
        Xn = np.reshape(X, (np.prod(S[0:n]), int(X.size / np.prod(S[0:n]))))

    else:  # the In x In+1...INI1...In-1 type
        if n == 1:
            Xn = np.reshape(X, (S[0], int(X.size / S[0])))
        elif n == N:
            Xn = np.reshape(X, (int(X.size / S[N - 1]), S[N - 1]))
            Xn = np.transpose(Xn, (1, 0))
        else:
            Xn = np.reshape(X, (np.prod(S[0:n - 1]), int(X.size / np.prod(S[0:n - 1]))))
            Xn = np.transpose(Xn, (1, 0))
            Xn = np.reshape(Xn, (S[n - 1], int(X.size / S[n - 1])))

    return Xn


'''
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% min: 1/2*||Z-X||^2 + ||X||_tr
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% [S, V, D, Sigma2] = MySVDtau(Z, tau);
% V = max(diag(V) - tau, 0);
% n = sum(V > 0);
% X = S(:, 1:n) * diag(V(1:n)) * D(:, 1:n)';

%% new
'''

def get_max_value(arr):
    m, n = np.shape(arr)
    max_matlab = np.zeros(n, dtype=type(arr[1][1]))
    for i in range(n):
        max_col = np.max(arr[:, i])
        max_matlab[i] = max_col
    return max_matlab


def Pro2TraceNorm(Z, tau):
    (m, n) = Z.shape
    if 2 * m < n:
        AAT = np.dot(Z, Z.T)
        S, Sigma2, D = np.linalg.svd(AAT)

        V = np.sqrt(Sigma2)

        tol = max(Z.shape) * np.finfo(max(V)).eps
        n = sum(V > max(tol, tau))

        mid = np.maximum(V[0:n - 1] - tau, 0) / V[0: n - 1]

        X = np.dot(np.dot(np.dot(S[:, 0:n - 1], np.diag(mid)), S[:, 0:n - 1].T), Z)

    elif m > 2 * n:
        X, n, Sigma2 = Pro2TraceNorm(Z.T, tau)
        X = X.T
    else:
        S, V, D = np.linalg.svd(Z)
        Sigma2 = V ** 2
        n = sum(V > tau)
        X = np.dot(np.dot(S[:, 0:n - 1], np.maximum(V[0:n - 1, 0:n - 1] - tau, 0)), D[:, 0:n - 1].T)

    return X, n, Sigma2


def RSE_fun(X, X_hat, W):
    # global RSE
    RSE_list = [0, 0, 0]
    err = X_hat.T.flatten() - X.T.flatten()
    RSE_list[0] = np.sqrt(np.sum(err ** 2) / np.sum((X.T.flatten()) ** 2))

    # RSE of observed entries
    X_hat_w = X_hat * W
    X_w = X * W
    err_w = X_hat_w.T.flatten() - X_w.T.flatten()
    RSE_list[1] = np.sqrt(np.sum(err_w ** 2) / np.sum((X_w.T.flatten()) ** 2))

    # RSE of missing entries
    Wr = np.logical_not(W)
    X_hat_wr = X_hat * Wr
    X_wr = X * Wr
    err_wr = X_hat_wr.T.flatten() - X_wr.T.flatten()
    RSE_list[2] = np.sqrt(np.sum(err_wr ** 2) / np.sum((X_wr.T.flatten()) ** 2))

    return RSE_list


# this is for simulation data, randomly generate tensor cores Z
def TR_initcoreten(S, r):
    N = int(np.size(S))
    Z = np.zeros((N,), dtype=np.object)  # 设置 dtype=np.object，可以在矩阵中设置形状不同的子矩阵
    for i in range(N - 1):
        Z[i] = np.random.randn(r[i], S[i], r[i + 1])
    Z[N - 1] = np.random.randn(r[N - 1], S[N - 1], r[0])

    return Z


import numpy as np
from math import prod
from trdstest.ops import my_chop2, factor, TR_initcoreten, circshift, tuple_ops, Z_neq, tenmat_sb
from trdstest.base import  tensor2mat, mat2tensor, init_tr_cores

def TRSVD(T, ep=1e-6):
    """returns all irreducible factors of x in vector
    T : ndarray
        d-dimensional tensor T
    ep : float
        prescribed relative error εp.
    Returns
    -------
    T : node
        Cores Zk
    r : 1D-array
        TR-ranks
    """
    n = T.shape
    d = len(n)
    node = {}
    r = np.ones((d))
    ep = ep / np.sqrt(d)

    for i in range(d - 1):
        T_ = T
        if i == 0:
            T = np.reshape(T_, (n[i], int(np.size(T_) / n[i])))
            u, s, v = np.linalg.svd(T, full_matrices=False)
            v = v.T
            # s = np.diag(s_)
            rc = my_chop2(s, np.sqrt(2) * ep * np.linalg.norm(s, 2))
            temp = np.cumprod(factor(rc))
            idx = int(np.abs(temp - np.sqrt(rc)) - 1)
            r[i + 1] = temp[idx]
            r[i] = rc / r[i + 1]
            u = u[:, 0:int(r[i] * r[i + 1])]
            u = np.reshape(u, (n[i], int(r[i + 1]), int(r[i])))
            node[i] = np.transpose(u, (2, 0, 1))
            s = s[0:int(r[i] * r[i + 1])]
            v = v[:, 0:int(r[i] * r[i + 1])]
            v = np.dot(v, np.diag(s)).T
            v = np.reshape(v, (int(r[i + 1]), int(r[i]), np.prod(n[1:])))
            T = np.transpose(v, (0, 2, 1))
        else:
            m = int(r[i] * n[i])
            T = np.reshape(T_, (m, int(np.size(T_) / m)))
            u, s, v = np.linalg.svd(T, full_matrices=False)
            v = v.T
            r1 = my_chop2(s, ep * np.linalg.norm(s, 2))
            r[i + 1] = np.maximum(r1, 1)
            u = u[:, 0:int(r[i + 1])]
            node[i] = np.reshape(u, (int(r[i]), n[i], int(r[i + 1])))
            v = v[:, 0:int(r[i + 1])]
            s = s[0:int(r[i + 1])]
            T = np.dot(v, np.diag(s))

    node[d - 1] = np.reshape(T, (int(r[d - 1]), n[d - 1], int(r[0])))

    return node, r



def tr_als(T, r, maxiter=10):
    """returns all irreducible factors of x in vector
    T : ndarray
        d-dimensional tensor T
    r : 1D-array
        predefined TR-ranks r.
    Returns
    -------
    node : 1D-array
        Cores Zk
    """
    tensor_size = T.shape
    dim = len(tensor_size)
    node = init_tr_cores(tensor_size, r)
    for i in range(maxiter):
        for n in range(1, dim+1):
            core_merge_flatten_trans = np.transpose(tensor2mat(core_merge(node, n), 2))
            G_neq_pinv = np.linalg.pinv(core_merge_flatten_trans)
            node[n-1] = mat2tensor(np.dot(tensor2mat(T, n), G_neq_pinv),2,node[n-1].shape)
    return node


# merge tr_cores EXCEPT the nth core
def core_merge(tr_cores, n):
    dim = len(tr_cores)
    tr_cores_shift = tr_cores[n:dim] + tr_cores[0:n] # shift the nth core to the last
    tr_mul = np.copy(tr_cores_shift[0])
    for i in range(dim-2):
        temp_core = np.copy(tr_cores_shift[i+1])
        zl = tr_mul.reshape(int(tr_mul.size/temp_core.shape[0]), temp_core.shape[0],  order='F').copy()
        zr = temp_core.reshape(temp_core.shape[0], temp_core.shape[1] * temp_core.shape[2],  order='F').copy()
        tr_mul = np.dot(zl, zr)
    s1 = tr_cores_shift[0].shape[0]
    s2 = tr_cores_shift[dim-2].shape[2]
    merge_neq_out = tr_mul.reshape(s1, int(tr_mul.size/(s1 * s2)), s2,  order='F').copy()
    return merge_neq_out


    # node, d, n, r = TRALS(T, ranks)

def TRALSAR(data, ep=1e-6):
    c = data
    switch = 0
    n = c.shape
    n = sorted(n)
    # n = n.flatten()
    d = len(n)
    ratio = 0.01 / d
    maxit = 20

    r = [1 for i in range(d)]
    node = []
    for i in range(d - 1):
        node.append(np.random.randn(int(r[i]), int(n[i]), int(r[i + 1])).reshape(1, 1, -1))
    node.append(np.random.randn(int(r[d - 1]), int(n[d - 1]), int(r[0])).reshape(1, 1, -1))

    od = []
    for ods in range(d):
        od.append(ods)

    for it in range(maxit * d):
        if it > 0:
            np.moveaxis(c, 0, -1)
            od = circshift(od)
        c = c.reshape(n[od[0]], c.size // n[od[0]], order='F')

        b = node[od[1]]
        for k0 in range(2, d):
            j0 = od[k0]
            br0 = node[j0]
            br0 = br0.reshape(r[j0], int(br0.size // r[j0]))
            b = b.reshape(int(b.size // r[j0]), r[j0])
            b = b @ br0
        # print('n', n)
        # print('r', r)
        # print('od', od)
        b = b.reshape((r[od[0]], r[od[1]], prod(tuple_ops(n, od[1:]))), order='F')
        b = b.transpose((0, 2, 1))
        b = b.reshape((r[od[1]] * r[od[0]], prod(tuple_ops(n, od[1:]))), order='F')
        a0 = np.linalg.lstsq(b.T, c.T, rcond=None)[0].T
        err0 = np.linalg.norm(c - a0 @ b) / np.linalg.norm(c.flatten())
        a0 = a0.reshape(r[od[0]], n[od[0]], r[od[1]])
        node[od[0]] = a0.transpose((2, 0, 1))

        r[od[1]] = r[od[1]] + 1
        node_od = np.mean(node[od[1]].flatten()) + np.std(node[od[1]].flatten()) * np.random.rand(n[od[1]], r[od[2]])
        node_od = node_od.reshape(np.size(node_od, 1), 1, -1)

        temp = node[od[1]].copy()
        if r[od[1]] > node[od[1]].shape[1]:
            node[od[1]] = np.zeros((np.size(node_od, 0), r[od[1]], np.size(node_od, 2)))
        for ii in range(np.size(node_od, 0)):
            node[od[1]][ii, 0:temp.shape[1], :] = temp[ii, 0:temp.shape[1], :]
            node[od[1]][ii, r[od[1]] - 1, :] = node_od[ii, :, :]
        # node[od[1]] = np.concatenate((node[od[1]], node_od), axis=0)

        b = node[od[1]].reshape(node[od[1]].shape[0] * node[od[1]].shape[1], node[od[1]].shape[2])

        for k in range(2, d):
            j1 = od[k]
            br1 = node[j1]
            br1 = br1.reshape(r[j1], int(br1.size / r[j1]))
            b = b.reshape(int(b.size / r[j1]), r[j1])
            b = b @ br1
        b = b.reshape((r[od[0]], r[od[1]], prod(tuple_ops(n, od[1:]))), order='F')
        b = b.transpose((0, 2, 1))
        b = b.reshape((r[od[1]] * r[od[0]], prod(tuple_ops(n, od[1:]))), order='F')
        a1 = np.linalg.lstsq(b.T, c.T, rcond=None)[0].T
        err1 = np.linalg.norm(c - a1 @ b) / np.linalg.norm(c.flatten())

        if (err0 - err1) / (err0) > ratio * (err0 - ep) / (err0) and err0 > ep:
            a1 = a1.reshape((r[od[0]], n[od[0]], r[od[1]]), order='F')
            node[od[0]] = a1.transpose((2, 0, 1))
            err0 = err1
            switch = 0
        else:
            temp = node[od[1]].copy()
            node[od[1]] = np.zeros((node[od[1]].shape[0], node[od[1]].shape[1] - 1, node[od[1]].shape[2]))
            temp1 = np.delete(temp, r[od[1]] - 1, axis=1)
            node[od[1]] = temp1
            r[od[1]] = r[od[1]] - 1
            switch = 1

        s = np.linalg.norm(node[od[0]].flatten())
        node[od[0]] = node[od[0]] / s
        print('it: %d, err=%f' % (it, err0))
        if err0 < ep and it >= 2 * d and switch == 1:
            break
        c = c.reshape(sorted(n, reverse=True), order='F')
    node[od[0]] = node[od[0]] * s

    return node, r




#!/usr/bin/env python
#-*- coding:utf-8 -*-
#############################################
# File Name: setup.py
# Author:
# Mail:
# Created Time:
#############################################
from setuptools import setup, find_packages
setup(
    name = "TRDstest",
    version = "0.1.7",
    keywords = ("pip", "tensor","tensor ring decomposition", "", ""),
    description = "tensor ring decomposition toolbox",
    long_description = "tensor ring decomposition toolbox",
    license = "BSD 3-Clause License",
    url = "",
    author = "",
    author_email = "",
    packages = find_packages(),
    include_package_data = True,
    platforms = "any",
    install_requires = []
)


import numpy as np
from math import gcd


def my_chop2(sv, eps):
    if eps <= 0.0:
        r = len(sv)
        return r
    sv0 = np.cumsum(abs(sv[::-1]) ** 2)[::-1]
    ff = [i for i in range(len(sv0)) if sv0[i] < eps ** 2]
    if len(ff) == 0:
        return len(sv)
    else:
        return np.amin(ff)

def factor(n):
    """returns all irreducible factors of x in vector
    n : int
    Returns
    -------
    T : vector
        factors
    """
    factors = []
    def get_factor(n):
        x_fixed = 2
        cycle_size = 2
        x = 2
        factor = 1
        while factor == 1:
            for count in range(cycle_size):
                if factor > 1: break
                x = (x * x + 1) % n
                factor = gcd(x - x_fixed, n)
            cycle_size *= 2
            x_fixed = x
        return factor
    while n > 1:
        next = get_factor(n)
        factors.append(next)
        n //= next
    return factors

# generate tr core tensors randomly
def init_tr_cores(tensor_size, tr_rank):
    tr_cores = []
    ndims = len(tensor_size)
    tr_rank.append(tr_rank[0])
    for n in range(0, ndims):
        tr_cores.append(0.1 * np.random.rand(tr_rank[n], tensor_size[n], tr_rank[n + 1]))
    return tr_cores

# backslash \
def backslash(a, b):
    m, n=a.shape
    p , q = b.shape
    if(m!=p):
        print('dimensions do not match!')
        return -1
    if(m==n):
        return np.linalg.solve(a, b)
    else:
        a1 = a.T.dot(a)
        b1 = a.T.dot(b)
        return np.linalg.solve(a1, b1)
# slash /
def slash(a,b):
    a=a.T
    b=b.T
    x = backslash(a,b)
    return x.T

def TR_initcoreten(T, r):
    '''

    :param T: narray
    :param r: list
    :return:
    Z: list
    '''
    S = T.shape
    N = int(np.size(S))
    # Z = np.zeros((N,), dtype=np.object)  # 设置 dtype=np.object，可以在矩阵中设置形状不同的子矩阵
    Z = []
    for i in range(N - 1):
        Z.append(np.random.randn(r[i], S[i], r[i + 1]))
    Z.append(np.random.randn(r[N - 1], S[N - 1], r[0]))

    return Z

def circshift(od):
    item = od.pop(0)
    od.append(item)
    return od

def tuple_ops(tpl, idxs):
    new_list = []
    for idx in idxs:
        new_list.append(tpl[idx])
    return new_list

def tenmat_sb(X, k):
    S = X.shape
    N = len(S)

    if k == 1:
        X_sb_k = np.reshape(X, (S[0], int(X.size / S[0])))
    elif k == N:
        X_sb_k = np.reshape(X, (int(X.size / S[N - 1]), S[N - 1]))
        X_sb_k = np.transpose(X_sb_k, (1, 0))

    else:
        X = np.reshape(X, (np.prod(S[0:k - 1]), int(X.size / np.prod(S[0:k - 1]))))
        X = np.transpose(X, [1, 0])
        X_sb_k = np.reshape(X, (S[k - 1], int(X.size / S[k - 1])))

    return X_sb_k


# formula 21 of <Tensor Ring Decomposition>
# Z is cell mode tensor cores
def TR_full_Z_k(Z, k):
    N = Z.size()


def Z_neq(Z, n):
    Z = np.roll(Z, -n - 1)  # arrange Z{n} to the last core, so we only need to multiply the first N-1 core
    N = np.size(Z, 0)
    P = Z[0]

    for i in range(N - 2):
        zl = np.reshape(P, (int(P.size / (np.size(Z[i], 2))), np.size(Z[i], 2)))
        zr = np.reshape(Z[i + 1], (np.size(Z[i + 1], 0), int(Z[i + 1].size / (np.size(Z[i + 1], 0)))))
        P = np.dot(zl, zr)
    Z_neq_out = np.reshape(P, (
    np.size(Z[0], 0), int(P.size / (np.size(Z[0], 0) * (np.size(Z[N - 1], 2)))), np.size(Z[N - 1], 2)))

    return Z_neq_out
